(* Dynamics *)
(* implements reduction on configuration where all possible
 * steps are taken in each round
 *)

structure Dynamics =
struct

structure I = IntSyn

datatype config = Config of I.proc list * int
(* Config([P1,...,Pn], n) where n is the next unused channel number *)

(* should not arise for well-typed configurations *)
exception DynamicError of string

type config = I.proc list

fun opt2list (NONE) = []
  | opt2list (SOME(P)) = [P]

fun select k ((l,P)::branches) = if k = l then P else select k branches
  | select k nil = raise DynamicError "label not recognized"

fun pp_proc (I.Send(x, msg, _)) = "send " ^ I.pp_chan x ^ " " ^ I.pp_msg msg ^ " ..."
  | pp_proc (I.Recv(x, match)) = "recv " ^ I.pp_chan x ^ " ..."
  | pp_proc (I.Fwd(x, y)) = "fwd " ^ I.pp_chan x ^ " " ^ I.pp_chan y
  | pp_proc (I.Call(p, x, ys)) = "call " ^ p ^ " " ^ I.pp_chan x ^ " " ^ String.concatWith " " (List.map I.pp_chan ys)
  | pp_proc (I.Cut(x, A_opt, P, Q)) = I.pp_chan x ^ " <- ... ; ..."
  | pp_proc (I.Marked(marked_P)) = pp_proc (Mark.data marked_P)

fun pp_config C = String.concatWith "\n" (List.map pp_proc C) ^ "\n% -----\n"

fun reduce (I.Label(k)) (I.MatchLabel(branches)) = select k branches
  | reduce (I.Unit) (I.MatchUnit(P)) = P
  | reduce (I.Channel(a)) (I.MatchChannel(x,P)) = I.subst_proc [(a,x)] P
  | reduce msg match = raise DynamicError "message / match mismatch"

(* split a (C0 @ [P] @ C2) C1rev = SOME(rev C0 @ C1rev, P, C2)
 * where P communicates along a
 * = NONE, if now such process P exists
 *)
fun split a (I.Send(b,msg,Popt)::C2) C1rev =
    if a = b then SOME(C1rev, I.Send(b,msg,Popt), C2)
    else split a C2 (I.Send(b,msg,Popt)::C1rev)
  | split a (I.Recv(b,match)::C2) C1rev =
    if a = b then SOME(C1rev, I.Recv(b,match), C2)
    else split a C2 (I.Recv(b,match)::C1rev)
  | split a (I.Marked(marked_P)::C2) C1rev =
    split a (Mark.data marked_P::C2) C1rev
  | split a (P::C2) C1rev = split a C2 (P::C1rev) (* ignore P = fwd, call, cut *)
  | split a nil C1rev = NONE

(* keep the state in reverse order *)
(* step_par env n changed C C' = (n', changed', rev D @ C') if C --> D in a parallel step
 *
 * The input C is already "reversed" with the "rightmost" process in a configuration
 * according to configuration typing comes first, so it is easy to lookup up
 * processes in the remainder of the input configuration.
 *
 * n is the next available free channel index in C
 * n' is the next available free channel index in rev D @ C'
 * changed indicates if a change reduction took place already in the parallel reduction
 * changed' is true if either changed is true or D is different from C
 *)

fun step_par env n changed (I.Recv(a,match)::C) C' =  (* positive: receive from the left *)
    (case split a C nil
      of SOME(C1rev, I.Send(a,msg,Popt), C2) => (* P :: a; move "right" *)
         step_par env n true (List.revAppend(C1rev, C2)) (opt2list Popt @ [reduce msg match] @ C')
       | NONE => step_par env n changed C (I.Recv(a,match)::C'))

  | step_par env n changed (I.Send(a,msg,Popt)::C) C' = (* negative: send to left *)
    (case split a C nil
      of SOME(C1rev, I.Recv(a,match), C2) => (* P :: a; move "right" *)
         step_par env n true (List.revAppend(C1rev, C2)) ([reduce msg match] @ opt2list Popt @ C')
       | NONE => step_par env n changed C (I.Send(a,msg,Popt)::C'))

  | step_par env n changed (I.Fwd(a,b)::C) C' = (* or: substitute everywhere in config *)
    (case split b C nil
      of SOME(C1rev, P, C2) => (* P :: b, positive or negative *)
         step_par env n true (List.revAppend(C1rev, C2)) (I.subst_proc [(a,b)] P::C')
       | NONE => step_par env n changed C (I.Fwd(a,b)::C'))

  | step_par env n changed (I.Cut(x, _, P, Q)::C) C' =
    let val a = I.Chan(n)       (* fresh channel *)
    in step_par env (n+1) true C (I.subst_proc [(a,x)] P::I.subst_proc [(a,x)] Q::C') end

  | step_par env n changed (I.Call(p, a, bs)::C) C' =
    let val SOME(I.ProcDef(p, I.Tp(x,nA), yBs, P, ext)) = I.lookup_procname env p (* must be defined *)
        val ys = List.map (fn I.Tp(y,nB) => y) yBs
        val P' = I.subst_proc (ListPair.zip (a::bs, x::ys)) P
    in step_par env n true C (P'::C') end

  | step_par env n changed (I.Marked(marked_P)::C) C' =
    step_par env n changed (Mark.data marked_P::C) C'

  | step_par env n changed nil C' = (changed, Config(List.rev C', n))

(* iterate env (n, C) = (n', C')
 * if C --> C' in a single parallel step
 * n is the next available channel index in C
 * n' is the next available channel index in C'
 * in parallel reduction, C = C' is possible, in which case we terminate
 *)
fun iterate env (Config(C,n)) =
    let (* val () = print ("before:\n" ^ pp_config C) *)
        val (changed, Config(C',n')) = step_par env n false C nil
        (* val () = print ("after:\n" ^ pp_config C') *)
    in
        if changed then iterate env (Config(C',n'))
        else Config(C',n') (* C = C' *)
    end

fun pp_msg (I.Chan(i)) msg = Int.toString i ^ " = " ^ I.pp_msg msg

(* observe env a [b1,...,bn] (n,C) = str
 * observe_frontier env [b1,...,bn] (n,C) = str
 * 
 * interact with the configuration to receive messages
 * along channels a, b1, ..., bn until a negative type
 * is encountered or the channel is closed with the '()' message
 *
 * n is the next available channel index in C for fresh allocation
 *)
fun observe env a frontier (Config(C,n)) =
    ((* print (pp_config C) ; *)
     case split a C nil
      of SOME(C1rev, I.Recv(a,match), C2) => (* negative *)
         " --- \n" ^ observe_frontier env frontier (Config(List.revAppend(C1rev, C2),n))
       | SOME(C1rev, I.Send(a, msg as I.Unit, NONE), C2) =>
         I.pp_msg msg ^ "\n" ^ observe_frontier env frontier (Config(List.revAppend(C1rev, C2), n))
       | SOME(C1rev, I.Send(a, msg as I.Label _, SOME(P)), C2) =>
         I.pp_msg msg ^ "." ^ next env a frontier (Config(P::List.revAppend(C1rev, C2), n))
       | SOME(C1rev, I.Send(a, msg as I.Channel(b), SOME(P)), C2) =>
         I.pp_msg msg ^ "." ^ next env a (b::frontier) (Config(P::List.revAppend(C1rev, C2), n))
       | NONE => raise DynamicError ("channel " ^ I.pp_chan a ^ " undefined"))

and next env a frontier (Config(C,n)) =
    observe env a frontier (iterate env (Config(C,n)))
and observe_frontier env nil (Config(C,n)) = ""
  | observe_frontier env (b::bs) (Config(C,n)) =
    I.pp_chan b ^ " -> " ^ observe env b bs (iterate env (Config(C,n)))

(* exec env p = ()
 * prints the messages sequence in channels visible
 * at the interface to process p, which must be closed except
 * for the channel provided
 *)
fun exec env p =
    let val SOME(I.ProcDef(p, I.Tp(x,nA), nil, P, ext)) = I.lookup_procname env p (* must be defined *)
        val P' = I.subst_proc [(I.Chan(0), x)] P
        val final = observe_frontier env [I.Chan(0)] (Config([P'],1)) (* 0 already used *)
        val () = if !Flags.verbosity >= 1 then print (final) else ()
    in () end

end (* structure Dynamics *)
